#!/usr/bin/python

# Inspired by https://www.piddlerintheroot.com/barcode-scanner/
# https://www.raspberrypi.org/forums/viewtopic.php?f=45&t=55100
# from 'brechmos' - thank-you!

CHARMAP_LOWERCASE = {
    4: "a",
    5: "b",
    6: "c",
    7: "d",
    8: "e",
    9: "f",
    10: "g",
    11: "h",
    12: "i",
    13: "j",
    14: "k",
    15: "l",
    16: "m",
    17: "n",
    18: "o",
    19: "p",
    20: "q",
    21: "r",
    22: "s",
    23: "t",
    24: "u",
    25: "v",
    26: "w",
    27: "x",
    28: "y",
    29: "z",
    30: "1",
    31: "2",
    32: "3",
    33: "4",
    34: "5",
    35: "6",
    36: "7",
    37: "8",
    38: "9",
    39: "0",
    44: " ",
    45: "-",
    46: "=",
    47: "[",
    48: "]",
    49: "\\",
    51: ";",
    52: "'",
    53: "~",
    54: ",",
    55: ".",
    56: "/",
}
CHARMAP_UPPERCASE = {
    4: "A",
    5: "B",
    6: "C",
    7: "D",
    8: "E",
    9: "F",
    10: "G",
    11: "H",
    12: "I",
    13: "J",
    14: "K",
    15: "L",
    16: "M",
    17: "N",
    18: "O",
    19: "P",
    20: "Q",
    21: "R",
    22: "S",
    23: "T",
    24: "U",
    25: "V",
    26: "W",
    27: "X",
    28: "Y",
    29: "Z",
    30: "!",
    31: "@",
    32: "#",
    33: "$",
    34: "%",
    35: "^",
    36: "&",
    37: "*",
    38: "(",
    39: ")",
    44: " ",
    45: "_",
    46: "+",
    47: "{",
    48: "}",
    49: "|",
    51: ":",
    52: '"',
    53: "~",
    54: "<",
    55: ">",
    56: "?",
}
CR_CHAR = 40
SHIFT_CHAR = 2
ERROR_CHARACTER = "?"


class BarcodeReader:
    def __init__(self, device_path="/dev/hidraw0"):
        self.device_path = device_path
        self.f = open(self.device_path, "rb")

    def read_barcode(self):
        barcode_string_output = ""
        # barcode can have a 'shift' character; this switches the character set
        # from the lower to upper case variant for the next character only.
        CHARMAP = CHARMAP_LOWERCASE
        while True:
            # step through returned character codes, ignore zeroes
            for char_code in [element for element in self.f.read(8) if element > 0]:
                if char_code == CR_CHAR:
                    # all barcodes end with a carriage return
                    self.f.close()
                    return barcode_string_output
                if char_code == SHIFT_CHAR:
                    # use uppercase character set next time
                    CHARMAP = CHARMAP_UPPERCASE
                else:
                    # if the charcode isn't recognized, use ?
                    barcode_string_output += CHARMAP.get(char_code, ERROR_CHARACTER)
                    # reset to lowercase character map
                    CHARMAP = CHARMAP_LOWERCASE


def barcode_reader(dev="/dev/hidraw0"):
    barcode_string_output = ""
    # barcode can have a 'shift' character; this switches the character set
    # from the lower to upper case variant for the next character only.
    CHARMAP = CHARMAP_LOWERCASE
    with open(dev, "rb") as fp:
        while True:
            # step through returned character codes, ignore zeroes
            for char_code in [element for element in fp.read(8) if element > 0]:
                if char_code == CR_CHAR:
                    # all barcodes end with a carriage return
                    return barcode_string_output
                if char_code == SHIFT_CHAR:
                    # use uppercase character set next time
                    CHARMAP = CHARMAP_UPPERCASE
                else:
                    # if the charcode isn't recognized, use ?
                    barcode_string_output += CHARMAP.get(char_code, ERROR_CHARACTER)
                    # reset to lowercase character map
                    CHARMAP = CHARMAP_LOWERCASE



import json

import click

from kort_client.api import KortAPI
from kort_client.run import PrintClient

from .config import settings


@click.group()
def cli():
    pass


@cli.command()
@click.argument("config_file", type=click.File("r"))
def setup(config_file):
    config = json.load(config_file)

    if settings.client_id:
        overwrite = click.confirm(
            "Client ID already set. Do you want to overwrite it?", default=False
        )

        if not overwrite:
            return

    print(config)
    if (
        "client_id" not in config
        or "client_secret" not in config
        or "base_url" not in config
    ):
        click.echo("Invalid config file")
        return

    settings.client_id = config["client_id"]
    settings.client_secret = config["client_secret"]
    settings.base_url = config["base_url"]

    settings.save_settings()

    api = KortAPI()

    printer_id = api.get_printer_id()

    api.set_printer_status(printer_id, "offline", "Newly registered printer")

    click.echo("Successfully registered printer")


@cli.command()
def config():
    click.echo("client_id: " + settings.get("client_id", ""))
    click.echo("client_secret: " + settings.get("client_secret", ""))
    click.echo("base_url: " + settings.get("base_url", ""))


@cli.command()
def run():
    click.echo("kort-client is running ...")

    api = KortAPI()
    client = PrintClient(settings, api)

    client.run()


if __name__ == "__main__":
    cli()



import usersettings

settings = usersettings.Settings("org.aleksis.apps.kort.client")
settings.add_setting("client_id", str, default="")
settings.add_setting("client_secret", str, default="")
settings.add_setting("base_url", str, default="")
settings.load_settings()



import os
import subprocess  # noqa
import time
from binascii import hexlify

import click
from nfc import ContactlessFrontend

from .usb_barcode_scanner.scanner import BarcodeReader


class CardDetector:
    id_ = None
    interactive = False

    def read_id(self):
        raise NotImplementedError()

    def abort(self):
        pass

    def clear(self):
        pass

    @classmethod
    def get_id(cls):
        if not cls.id_:
            raise NotImplementedError()
        return cls.id_

    @classmethod
    def get_card_detectors(cls):
        card_detectors = cls.__subclasses__()
        return {d.get_id(): d() for d in card_detectors}


class EvolisCardDetector(CardDetector):
    """This is a card detector for Evolis card printers like Evolis Primacy.

    To use it, you have to get the evocom tool from the Evolis customer support.
    Then place it at /opt/evocom/evocom.
    """

    id_ = "evolis_evocom"
    EVOCOM_PATH = "/opt/evocom/evocom"

    def get_detector(self):
        from kort_client.run import KortClientException

        files = []
        for __, __, files in os.walk("/dev/"):
            files = files
            break
        files = {f for f in files if f.startswith("hidraw")}
        possible_interfaces = []
        for hid_interface in list(files):
            path_to_read = f"/sys/class/hidraw/{hid_interface}/device/uevent"
            with open(path_to_read) as f:
                content = f.read()
                if "rfid" in content.lower():
                    possible_interfaces.append(hid_interface)

        if len(possible_interfaces) != 1:
            raise KortClientException("No or more than one  reader found")

        return os.path.join("/dev", possible_interfaces[0])

    def get_printer(self):
        from kort_client.run import KortClientException

        files = []
        for __, __, files in os.walk("/dev/usb/"):
            files += files

        files = {f for f in files if f.startswith("lp")}

        if len(files) != 1:
            raise KortClientException("No or more than one printer found")

        suffix = list(files)[0]
        return f"/dev/usb/{suffix}"

    def read_id(self):
        from kort_client.run import KortClientException

        reader = BarcodeReader(self.get_detector())
        # Transport the card to the printer
        try:
            r = subprocess.run(  # noqa
                [self.EVOCOM_PATH, "-p", self.get_printer(), "Sis"]
            )
        except (subprocess.CalledProcessError, OSError) as e:
            raise KortClientException(e)

        # Read the card
        chip_number = reader.read_barcode()

        try:
            r.check_returncode()
        except (subprocess.CalledProcessError, OSError) as e:
            raise KortClientException(e)
        return chip_number

    def clear(self):
        subprocess.run([self.EVOCOM_PATH, "-p", self.get_printer(), "Ser"])  # noqa

    def abort(self):
        self.clear()


class NfcReadUidDetector(CardDetector):
    id_ = "nfc_read_uid"

    def read_id(self):
        reader = ContactlessFrontend()
        reader.open("usb")
        from kort_client.run import KortClientException

        try:
            click.secho("→ Please hold your card to the reader")
            tag = reader.connect(rdwr={"on-connect": lambda tag: False})
            uid = hexlify(tag._nfcid).decode("utf-8")
            click.secho("Card successfully detected", fg="green")
            click.secho(f"UID: {uid}", fg="white")

            click.secho("Please remove the card from the reader", fg="red")
            while True:
                started = time.time()
                after5s = lambda: time.time() - started > 0.5
                tag = reader.connect(
                    rdwr={"on-connect": lambda tag: False}, terminate=after5s
                )
                if not tag:
                    break
                new_uid = hexlify(tag._nfcid).decode("utf-8")
                if uid != new_uid:
                    break
            click.secho("The card has been removed.", fg="green")

        except Exception as e:
            raise KortClientException(e.args)

        return uid



import urllib
from typing import Any, Optional, Sequence
from urllib.parse import urljoin

import click
import requests as requests
from requests import request

from .config import settings


class KortAPIException(Exception):
    pass


class ConfigurationException(KortAPIException):
    pass


class KortAPI:
    URLS = {
        "authorize": "oauth/authorize/",
        "token": "oauth/token/",
        "status": "app/kort/api/v1/printers/{}/status/",
        "printer": "app/kort/api/v1/printers/",
        "next_job": "app/kort/api/v1/printers/{}/jobs/next/",
        "job_status": "app/kort/api/v1/jobs/{}/status/",
        "chip_number": "app/kort/api/v1/jobs/{}/chip_number/",
    }

    def __init__(self):
        self.settings = settings
        self._validate_settings()

    def get_printer(self) -> dict[str, Any]:
        r = self._do_request("GET", "printer")
        return r

    def get_printer_id(self) -> int:
        r = self.get_printer()
        return r["id"]

    def get_next_job(self, printer_id: int) -> Optional[dict[str, Any]]:
        r = self._do_request("GET", "next_job", [printer_id])
        print(r)
        if r.get("status") == "no_job":
            return None
        return r

    def set_job_status(
        self, job_id: int, status: str, status_text: str = None
    ) -> dict[str, Any]:
        data = {"status": status}
        if status_text is not None:
            data["status_text"] = status_text
        r = self._do_request("PUT", "job_status", [job_id], data=data)
        return r

    def set_chip_number(self, job_id: int, chip_number: str) -> dict[str, Any]:
        data = {"chip_number": chip_number}
        r = self._do_request("PUT", "chip_number", [job_id], data=data)
        return r

    def set_printer_status(
        self, printer_id: int, status: str, status_text: str = None
    ) -> dict[str, Any]:
        data = {"status": status}
        if status_text is not None:
            data["status_text"] = status_text
        r = self._do_request("PUT", "status", [printer_id], data=data)
        return r

    def _validate_settings(self):
        if not self.settings.client_id:
            raise ConfigurationException("No client_id set.")
        if not self.settings.client_secret:
            raise ConfigurationException("No client_secret set.")
        if not self.settings.base_url:
            raise ConfigurationException("No base_url set.")

    def _do_request(
        self, method: str, url: str, attrs: Optional[Sequence] = None, **kwargs
    ) -> dict[str, Any]:
        if url in self.URLS:
            url = self.URLS[url]
        if attrs:
            url = url.format(*attrs)
        url = urljoin(settings.base_url, url)
        try:
            r = request(
                method,
                url,
                auth=(
                    urllib.parse.quote(self.settings.client_id),
                    urllib.parse.quote(self.settings.client_secret),
                ),
                **kwargs
            )
            if not r.ok:
                click.secho("Error {}: {}".format(r.status_code, r.text), fg="red")
                raise KortAPIException("Error {}: {}".format(r.status_code, r.text))
            return r.json()
        except (requests.RequestException, requests.JSONDecodeError) as e:
            click.secho("Error: " + str(e), fg="red")
            raise KortAPIException(e)



import tempfile
import time
from urllib.parse import urljoin

import click
import cups
import requests

from kort_client.api import KortAPIException
from kort_client.card_detectors import CardDetector


class KortClientException(Exception):
    pass


IPP_CODES = {y: x for x, y in cups.__dict__.items() if x.startswith("IPP_JOB")}


class PrintClient:
    def __init__(self, settings, api, interactive=False):
        self.settings = settings
        self.api = api
        self.interactive = interactive
        self.with_printer = True

        self.printer_id = None
        self.conn = cups.Connection()
        self.card_detectors = CardDetector.get_card_detectors()
        self.card_detector = None
        self.card_detector_in_process = False

    def run(self):
        try:
            while True:
                try:
                    self.step()
                except (KortAPIException, KortClientException) as e:
                    printer_status = "with_errors"
                    status_text = str(e)
                    click.secho(
                        "An error occurred, try again in five seconds: {}".format(e),
                        fg="red",
                    )
                    if self.card_detector_in_process and self.card_detector:
                        self.card_detector.abort()
                else:
                    printer_status = "online"
                    status_text = ""

                if self.printer_id:
                    # Set status
                    self.api.set_printer_status(
                        self.printer_id, printer_status, status_text
                    )

                if printer_status == "with_errors":
                    time.sleep(5)
                else:
                    time.sleep(1)

        except (EOFError, KeyboardInterrupt):
            if self.printer_id:
                click.echo("Set printer as offline")
                self.api.set_printer_status(
                    self.printer_id, "offline", "Printer client was stopped by user."
                )
            if self.card_detector and self.card_detector_in_process:
                self.card_detector.abort()
            raise

    def _validate_printer(self):
        self.with_printer = self.printer_config.get("cups_printer")
        self.card_detector = self.card_detectors.get(
            self.printer_config["card_detector"], None
        )

        if not self.with_printer and not self.card_detector:
            raise KortClientException(
                "No printer or card detector configured for printer {}".format(
                    self.printer_config["name"]
                )
            )

        if self.with_printer:
            cups_printers = self.conn.getPrinters()
            if self.printer_config["cups_printer"] not in cups_printers:
                raise KortClientException(
                    "CUPS printer {} not found".format(
                        self.printer_config["cups_printer"]
                    ),
                )

    def step(self):
        self.printer_config = self.api.get_printer()
        print(self.printer_config)

        self.printer_id = self.printer_config["id"]

        self._validate_printer()

        self.card_detector_in_process = False
        if self.card_detector:
            self.interactive = self.card_detector.interactive
            self.card_detector.clear()

        next_job = self.api.get_next_job(self.printer_id)

        print(next_job)
        if next_job:
            job_id = next_job["id"]
            click.secho("Got new print job {}".format(next_job), fg="green")

            self.api.set_job_status(
                job_id,
                "in_progress",
            )

            if not self.printer_config.get("generate_number_on_server"):
                # Now do something to set the number, but irrelevant for this example
                click.secho("Generate number on server disabled", fg="yellow")
                if self.card_detector:
                    self.card_detector_in_process = True
                    chip_number = self.card_detector.read_id()
                    if chip_number:
                        next_job = self.api.set_chip_number(job_id, str(chip_number))
                        time.sleep(1)
                    else:
                        raise KortClientException("No valid chip number read")
                else:
                    raise KortClientException("No card detector configured")

            if self.with_printer and next_job["card"]["chip_number"]:
                # Download PDF file
                with tempfile.NamedTemporaryFile("wb") as f:
                    r = requests.get(
                        urljoin(self.settings.base_url, next_job["card"]["pdf_file"])
                    )
                    f.write(r.content)
                    f.flush()

                    cups_job_id = self.conn.printFile(
                        self.printer_config["cups_printer"],
                        f.name,
                        f"AlekSIS: #{job_id}",
                        {},
                    )

                last_job_state = None
                while True:
                    a = self.conn.getJobAttributes(cups_job_id)
                    current_job_state = a["job-state"]
                    status = "in_progress"

                    if current_job_state in (
                        cups.IPP_JOB_STOPPED,
                        cups.IPP_JOB_CANCELED,
                        cups.IPP_JOB_ABORTED,
                        cups.IPP_JOB_HELD,
                    ):
                        status = "failed"
                    elif current_job_state == cups.IPP_JOB_COMPLETED:
                        status = "finished"

                    if last_job_state != current_job_state:
                        click.echo(
                            click.style(
                                f"Print job is in state {status}: {current_job_state}",
                                fg="yellow",
                            )
                        )
                        last_job_state = current_job_state
                        self.api.set_job_status(
                            job_id,
                            status,
                            IPP_CODES.get(current_job_state, str(current_job_state)),
                        )

                    if status in ("failed", "finished"):
                        break

                    time.sleep(0.5)
